<?php

declare(strict_types=1);

namespace Drupal\otlog\Logger;

use Drupal\Core\Logger\LogMessageParserInterface;
use Drupal\Core\Logger\RfcLogLevel;
use Drupal\Core\Logger\RfcLoggerTrait;
use Drupal\Core\Utility\Error;
use OpenTelemetry\API\Globals;
use OpenTelemetry\API\Logs\LogRecord;
use OpenTelemetry\API\Logs\LoggerProviderInterface;
use OpenTelemetry\API\Logs\Severity;
use OpenTelemetry\SemConv\TraceAttributes;
use Psr\Log\LogLevel;
use Psr\Log\LoggerInterface;

/**
 * Logs events to an OpenTelemetry log endpoint.
 *
 * @see https://opentelemetry.io/docs/specs/otel/logs/bridge-api/
 */
class OtLog implements LoggerInterface {

  use RfcLoggerTrait;

  /**
   * Constructs an OtLog object.
   */
  public function __construct(
    protected readonly LogMessageParserInterface $parser,
    protected readonly ?LoggerProviderInterface $loggerProvider = NULL,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function log($level, string|\Stringable $message, array $context = []): void {
    $logger = $this->loggerProvider()->getLogger('org.drupal.otlog');
    if ($logger->isEnabled()) {
      $attributes = [
        TraceAttributes::USER_ID => $context['uid'],
        TraceAttributes::URL_FULL => $context['request_uri'],
        TraceAttributes::CLIENT_ADDRESS => $context['ip'],
      ];

      // Extract error attributes from error and exceptions messages.
      if (str_contains((string) $message, Error::DEFAULT_ERROR_MESSAGE)) {
        $attributes += [
          TraceAttributes::ERROR_TYPE => $context['%type'],
          TraceAttributes::CODE_FUNCTION => $context['%function'],
          TraceAttributes::CODE_FILEPATH => $context['%file'],
          TraceAttributes::CODE_LINENO => $context['%line'],
        ];
      }

      // Extract backtrace into attributes and remove it from the log context.
      if (isset($context['@backtrace_string'])) {
        $attributes[TraceAttributes::CODE_STACKTRACE] = $context['@backtrace_string'];
        unset($context['@backtrace_string']);
      }

      // Cast message to string.
      $message = (string) $message;

      // Convert PSR3-style message and placeholders. Suppress backtrace
      // placeholder.
      $placeholders = $this->parser->parseMessagePlaceholders($message, $context);
      $placeholders['@backtrace_string'] = '';

      // Render placeholders into message body without translation.
      $body = trim(strtr($message, $placeholders));

      $severityText = $this->severityText($level);
      $record = (new LogRecord("[{$context['channel']}] [{$severityText}] {$body}"))
        ->setSeverityNumber($this->severityNumber($level))
        ->setSeverityText($severityText)
        ->setTimestamp($context['timestamp'] * LogRecord::NANOS_PER_SECOND)
        ->setAttributes($attributes);
      $logger->emit($record);
    }
  }

  /**
   * Maps RFC 5424 log level to OpenTelemetry severity number.
   *
   * @see https://opentelemetry.io/docs/specs/otel/logs/data-model-appendix/#appendix-b-severitynumber-example-mappings
   */
  protected function severityNumber(mixed $level): Severity {
    return match ($level) {
      RfcLogLevel::EMERGENCY => Severity::FATAL,
      RfcLogLevel::ALERT => Severity::ERROR3,
      RfcLogLevel::CRITICAL => Severity::ERROR2,
      RfcLogLevel::ERROR => Severity::ERROR,
      RfcLogLevel::WARNING => Severity::WARN,
      RfcLogLevel::NOTICE => Severity::INFO2,
      RfcLogLevel::INFO => Severity::INFO,
      RfcLogLevel::DEBUG => Severity::DEBUG,
      default => Severity::WARN2,
    };
  }

  /**
   * Maps RFC 5424 log level to severity text.
   */
  protected function severityText(mixed $level): string {
    return match ($level) {
      RfcLogLevel::EMERGENCY => LogLevel::EMERGENCY,
      RfcLogLevel::ALERT => LogLevel::ALERT,
      RfcLogLevel::CRITICAL => LogLevel::CRITICAL,
      RfcLogLevel::ERROR => LogLevel::ERROR,
      RfcLogLevel::WARNING => LogLevel::WARNING,
      RfcLogLevel::NOTICE => LogLevel::NOTICE,
      RfcLogLevel::INFO => LogLevel::INFO,
      RfcLogLevel::DEBUG => LogLevel::DEBUG,
      default => LogLevel::WARNING,
    };
  }

  /**
   * Gets the logger provider.
   */
  protected function loggerProvider(): LoggerProviderInterface {
    return $this->loggerProvider ?? Globals::loggerProvider();
  }

}
